#ifndef TWINKLER_H
#define TWINKLER_H

#if defined _WIN32
	#include <windows.h>
#endif
#include <stdlib.h>
#include "ftd2xx.h"

enum {
    Twinkler_ColourArray = 0xF0,
    Twinkler_SetColour,
    Twinkler_FadePeriod,
    Twinkler_BlinkRate,
    Twinkler_SetState,
    Twinkler_SetRange,
    Twinkler_SetBaud,
    Twinkler_StoreState,
    Twinkler_TickMark,
};

#define Twinkler_RGB(red, green, blue)  \
      (unsigned char)(36 * (red) + 6 * (green) + (blue))



#if defined __cplusplus

class TwinklerIO {
    /* abstract base class */
public:
    virtual ~TwinklerIO() {}
    virtual bool OpenPort(const char *PortName, unsigned BaudRate) = 0;
    virtual void ClosePort() = 0;
    virtual bool SendData(const unsigned char *Data, size_t DataSize) = 0;
    virtual bool SendBreak() = 0;
    virtual unsigned char BitBang(unsigned char Mask, unsigned char Bits) = 0;
};


class TwinklerCom : public TwinklerIO {

private:
    #if defined _WIN32
        HANDLE hCom;
    #else /* _WIN32 */
        int fdCom;
        struct termios oldtio;
    #endif /* _WIN32 */

public:
    TwinklerCom() {
        #if defined _WIN32
            hCom = INVALID_HANDLE_VALUE;
        #else /* _WIN32 */
            fdCom = -1;
        #endif /* _WIN32 */
    }

    virtual ~TwinklerCom()
    {
        ClosePort();
    }

    /* ----- low-level serial port routines ----- */

    virtual bool OpenPort(const char *PortName, unsigned BaudRate)
    {
        #if defined _WIN32
            DCB dcb;

            hCom = CreateFile(PortName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
            if (hCom == INVALID_HANDLE_VALUE)
                return false;

            GetCommState(hCom, &dcb);
            dcb.BaudRate = BaudRate;
            dcb.ByteSize = 8;
            dcb.StopBits = ONESTOPBIT;
            dcb.Parity = NOPARITY;
            dcb.fDtrControl = DTR_CONTROL_DISABLE;
            dcb.fOutX = FALSE;
            dcb.fInX = FALSE;
            dcb.fNull = FALSE;
            dcb.fRtsControl = RTS_CONTROL_DISABLE;
            SetCommState(hCom, &dcb);
        #else /* _WIN32 */
            struct termios newtio;
            char portdev[60];

            fdCom = open(port, O_RDWR | O_NOCTTY | O_NONBLOCK | O_NDELAY);
            tcgetattr(fdCom, &oldtio); /* save current port settings */
            memset(&newtio, 0, sizeof newtio);

            newtio.c_cflag = CS8 | CLOCAL | CREAD;
            switch (BaudRate) {
            #ifdef B115200
                case  115200: newtio.c_cflag |= B115200; break;
            #endif
            #ifdef B57600
                case   57600: newtio.c_cflag |= B57600;  break;
            #endif
            #ifdef B38400
                case   38400: newtio.c_cflag |= B38400;  break;
            #endif
            #ifdef B19200
                case   19200: newtio.c_cflag |= B19200;  break;
            #endif
            #ifdef B9600
                case    9600: newtio.c_cflag |= B9600;   break;
            #endif
            default:
                ClosePort();
                return false;
            } /* switch */

            newtio.c_iflag = IGNPAR | IGNBRK; /* ignore parity and BREAK conditions */
            newtio.c_oflag = 0; /* set output mode (non-canonical, no processing,...) */
            newtio.c_lflag = 0; /* set input mode (non-canonical, no echo,...) */

            tcflush(fdCom, TCIFLUSH);
            if (tcsetattr(fdCom, TCSANOW, &newtio))
                return false;
        #endif /* _WIN32 */

        return true;
    }

    virtual void ClosePort()
    {
        #if defined _WIN32
            if (hCom != INVALID_HANDLE_VALUE) {
                FlushFileBuffers(hCom);
                CloseHandle(hCom);
                hCom = INVALID_HANDLE_VALUE;
            } /* if */
        #else /* _WIN32 */
            if (fdCom >= 0) {
                tcflush(fdCom, TCOFLUSH);
                tcflush(fdCom, TCIFLUSH);
                tcsetattr(fdCom, TCSANOW, &oldtio);
                close(fdCom);
                fdCom = -1;
            } /* if */
        #endif /* _WIN32 */
    }

    virtual bool SendData(const unsigned char *Data, size_t DataSize)
    {
        if (!isComOpen())
            return false;
        #if defined _WIN32
            DWORD written = 0;
            if (!WriteFile(hCom, Data, DataSize, &written, NULL))
                written = 0;
            return written == DataSize;
        #else /* _WIN32 */
            return write(fdCom, Data, DataSize) == DataSize;
        #endif /* _WIN32 */
    }

    virtual bool SendBreak()
    {
        if (!isComOpen())
            return false;
        #if defined _WIN32
            SetCommBreak(hCom);
            Sleep(200);
            ClearCommBreak(hCom);
        #else
            tcsendbreak(fdCom, 0);
        #endif
        return true;
    }

    virtual unsigned char BitBang(unsigned char, unsigned char)
    {
        return 0xff;
    }

    /* ----- helper functions ----- */
private:
    bool isComOpen()
    {
        #if defined _WIN32
            return hCom!=INVALID_HANDLE_VALUE;
        #else /* _WIN32 */
            return fdCom >= 0;
        #endif /* _WIN32 */
    }
};


class TwinklerNative: public TwinklerIO {
private:
    FT_HANDLE hFT;

public:
    TwinklerNative()
    {
        hFT = INVALID_HANDLE_VALUE;
    }

    virtual ~TwinklerNative()
    {
        ClosePort();
    }

    virtual bool OpenPort(const char *PortName, unsigned BaudRate)
    {
        int index = atoi(PortName);
        if (FT_Open(index, &hFT) == FT_OK) {
            FT_ResetDevice(hFT);
            FT_SetBaudRate(hFT, BaudRate);
            FT_SetDataCharacteristics(hFT, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
            FT_SetFlowControl(hFT, FT_FLOW_NONE, 0, 0);
        } else {
            hFT = INVALID_HANDLE_VALUE;
        } /* if */
        return hFT != INVALID_HANDLE_VALUE;
    }

    virtual void ClosePort()
    {
        if (hFT != INVALID_HANDLE_VALUE) {
            FT_Close(hFT);
            hFT = INVALID_HANDLE_VALUE;
        } /* if */
    }

    virtual bool SendData(const unsigned char *Data, size_t DataSize)
    {
        DWORD written = 0;
        if (hFT == INVALID_HANDLE_VALUE || FT_Write(hFT, (void*)Data, DataSize, &written) != FT_OK)
            written = 0;
        return written == DataSize;
    }

    virtual bool SendBreak()
    {
        if (hFT == INVALID_HANDLE_VALUE)
            return false;
        FT_SetBreakOn(hFT);
        #if defined _WIN32
            Sleep(200);
        #else
            /* for Linux */
            usleep(200*1000);
        #endif
        FT_SetBreakOff(hFT);
        return true;
    }

    virtual unsigned char BitBang(unsigned char Mask, unsigned char Bits)
    {
        if (hFT == INVALID_HANDLE_VALUE)
            return 0;
        unsigned char result;
        unsigned char msk = (unsigned char)((Mask << 4) | Bits);
        FT_SetBitMode(hFT, msk, 0x20);
        FT_GetBitMode(hFT, &result);
        return (unsigned char)(result & 0x0f);
    }
};


class Twinkler {
private:
    TwinklerIO *Port;
    unsigned char cbus;

public:
    /* ----- constructor/ destructor ----- */
    Twinkler(TwinklerIO *port) : Port(port) {}

    ~Twinkler()
    {
        Port->ClosePort();
    }

    /* ----- forwarded low-level functions ----- */

    bool OpenPort(const char *PortName, unsigned BaudRate)
    {
        return Port->OpenPort(PortName, BaudRate);
    }

    void ClosePort()
    {
        Port->ClosePort();
    }

    bool SendData(const unsigned char *Data, size_t DataSize)
    {
        return Port->SendData(Data, DataSize);
    }

    bool Reset()
    {
        return Port->SendBreak();
    }

    unsigned char BitBang(unsigned char Mask, unsigned char Bits)
    {
        return Port->BitBang(Mask, Bits);
    }


    /* ----- high-level functions ----- */

    bool ColourArray(const unsigned char *Data, size_t DataSize)
    {
        unsigned char cmd = Twinkler_ColourArray;
        Port->SendData(&cmd, 1);
        return Port->SendData(Data, DataSize);
    }

    bool SetColour(unsigned char Colour)
    {
        unsigned char cmd[2] = { Twinkler_SetColour, Colour };
        return Port->SendData(cmd, 2);
    }

    bool SetColour(unsigned char Red, unsigned char Green, unsigned char Blue)
    {
        unsigned char cmd[2] = { Twinkler_SetColour, Twinkler_RGB(Red, Green, Blue) };
        return Port->SendData(cmd, 2);
    }

    bool FadePeriod(unsigned char Period)
    {
        unsigned char cmd[2] = { Twinkler_FadePeriod, Period };
        return Port->SendData(cmd, 2);
    }

    bool BlinkRate(unsigned char Rate)
    {
        unsigned char cmd[2] = { Twinkler_BlinkRate, Rate };
        return Port->SendData(cmd, 2);
    }

    bool SetState(unsigned char State)
    {
        unsigned char cmd[2] = { Twinkler_SetState, State };
        return Port->SendData(cmd, 2);
    }

    bool SetRange(int First = 0, int Last = 16255)
    {
        unsigned char cmd[5] = { Twinkler_SetRange };
        cmd[1] = (unsigned char)(First & 0x7f);
        cmd[2] = (unsigned char)((First >> 7) & 0x7e);
        cmd[3] = (unsigned char)((Last - First) & 0x7f);
        cmd[4] = (unsigned char)(((Last - First) >> 7) & 0x7e);
        return Port->SendData(cmd, 5);
    }

    bool SetBaud(unsigned char Baud)
    {
        unsigned char cmd[2] = { Twinkler_SetBaud, Baud };
        return Port->SendData(cmd, 2);
    }

    bool StoreState(unsigned char Hold, unsigned char State)
    {
        unsigned char cmd[3] = { Twinkler_StoreState, Hold, State };
        return Port->SendData(cmd, 3);
    }

    bool TickMark()
    {
        unsigned char cmd = Twinkler_TickMark;
        return Port->SendData(&cmd, 1);
    }

    unsigned char ReadInputs()
    {
    	unsigned char sw = Port->BitBang(0, 0);
    	if (sw != cbus) {
    		/* debounce */
    		int count = 0;
    		while (count < 5) {
    			cbus = sw;
    			#if defined _WIN32
    				Sleep(5);
    			#else
    				usleep(5*1000);
    			#endif
    			sw = Port->BitBang(0, 0);
    			count = (cbus == sw) ? count + 1 : 0;
    		} /* while */
    	} /* if */
    	return sw;
    }
};

#endif /* __cplusplus */

#endif /* TWINKLER_H */
